import os
import json
from pathlib import Path

from nasbench301 import representations
from surrogate_models import utils
from surrogate_models.ensemble import Ensemble


fixed_hyperparameters = {
        "CreateImageDataLoader:batch_size": 96,
        "ImageAugmentation:augment": "True",
        "ImageAugmentation:cutout": "True",
        "ImageAugmentation:cutout_holes": 1,
        "ImageAugmentation:cutout_length": 16,
        "ImageAugmentation:autoaugment": "False",
        "ImageAugmentation:fastautoaugment": "False",
        "LossModuleSelectorIndices:loss_module": "cross_entropy",
        "NetworkSelectorDatasetInfo:darts:auxiliary": "True",
        "NetworkSelectorDatasetInfo:darts:drop_path_prob": 0.2,
        "NetworkSelectorDatasetInfo:network": "darts",
        "OptimizerSelector:optimizer": "sgd",
        "OptimizerSelector:sgd:learning_rate": 0.025,
        "OptimizerSelector:sgd:momentum": 0.9,
        "OptimizerSelector:sgd:weight_decay": 0.0003,
        "SimpleLearningrateSchedulerSelector:cosine_annealing:T_max": 100,
        "SimpleLearningrateSchedulerSelector:cosine_annealing:eta_min": 1e-8,
        "SimpleLearningrateSchedulerSelector:lr_scheduler": "cosine_annealing",
        "SimpleTrainNode:batch_loss_computation_technique": "mixup",
        "SimpleTrainNode:mixup:alpha": 0.2,
        "NetworkSelectorDatasetInfo:darts:init_channels": 32,
        "NetworkSelectorDatasetInfo:darts:layers": 8
        }


def load_ensemble(ensemble_parent_directory):
    """Loads a surrogate ensemble
    
    args:
        ensemble_parent_directory: directory which contains the ensemble members. Members must be the same model type
    """

    ensemble_member_dirs = [os.path.dirname(filename) for filename in Path(ensemble_parent_directory).rglob('*surrogate_model.model')]
    data_config = json.load(open(os.path.join(ensemble_member_dirs[0], 'data_config.json'), 'r'))
    model_config = json.load(open(os.path.join(ensemble_member_dirs[0], 'model_config.json'), 'r'))

    surrogate_model = Ensemble(member_model_name=model_config['model'],
                               data_root='None', 
                               log_dir=ensemble_parent_directory,
                               starting_seed=data_config["seed"],
                               model_config=model_config,
                               data_config=data_config, 
                               ensemble_size=len(ensemble_member_dirs),
                               init_ensemble=False)

    surrogate_model.load(model_paths=ensemble_member_dirs)

    surrogate_api = SurrogateAPI(surrogate_model)

    return surrogate_api


class SurrogateAPI():
    """Wrapper for a surrogate ensemble"""

    def __init__(self, surrogate_model):
        """
        args:
            surrogate_model: An instance of Ensemble
        """
        self.model = surrogate_model
        self.representations_converters = representations.CONVERTER_DICT

    def convert_representation(self, config, representation):
        """Convert representation to a dictionary"""
        if not representation in self.representations_converters.keys():
            raise NotImplementedError("%s representation is not supported, please choose from %s" %(representation, self.representations_converters.keys()))
        
        converter = self.representations_converters[representation]()
        config_dict = converter.convert(config)
        return config_dict

    def predict(self, config, representation, with_noise=True):
        """Return the mean over the predictions of surrogate ensemble individuals with or without noise
        
        args:
            config: An architecture to query, given in any of the supported representations
            representation: str, representation used for the config 
            with_noise: bool, wether to apply noise or only use the mean of the ensemble members as prediction
        """
        config_dict = self.convert_representation(config, representation)

        config_dict = {**fixed_hyperparameters, **config_dict}

        if with_noise:
            pred = self.model.query(config_dict)
        else:
            pred = self.model.query_mean(config_dict)
        return pred



from collections import namedtuple


OPS = [
        'max_pool_3x3',
        'avg_pool_3x3',
        'skip_connect',
        'sep_conv_3x3',
        'sep_conv_5x5',
        'dil_conv_3x3',
        'dil_conv_5x5'
       ]

Genotype = namedtuple('Genotype', 'normal normal_concat reduce reduce_concat')


def convert_genotype_to_config(arch):
    """Converts a DARTS genotype to a configspace instance dictionary"""
    base_string = 'NetworkSelectorDatasetInfo:darts:'
    config = {}

    for cell_type in ['normal', 'reduce']:
        cell = eval('arch.' + cell_type)

        start = 0
        n = 2
        for node_idx in range(4):
            end = start + n
            ops = cell[2 * node_idx: 2 * node_idx + 2]

            # get edge idx
            edges = {base_string + 'edge_' + cell_type + '_' + str(start + i): op for
                         op, i in ops}
            config.update(edges)

            if node_idx != 0:
                # get node idx
                input_nodes = sorted(list(map(lambda x: x[1], ops)))
                input_nodes_idx = '_'.join([str(i) for i in input_nodes])
                config.update({base_string + 'inputs_node_' + cell_type + '_' + str(node_idx + 2):
                                   input_nodes_idx})

            start = end
            n += 1
    return config


class BaseConverter():
    """Base class for converters from one representation to a dictionary"""

    def __init__(self, name):
        self.name = name

    def convert(self, config):
        raise NotImplementedError("Child classes have to implement convert.")


class ConfigspaceInstanceConverter(BaseConverter):
    """Converter for a ConfigSpace sample to dictionary. Does nothing if it receives a dict"""
    
    def __init__(self, name="configspace"):
        super().__init__(name)

    def convert(self, config):
        if isinstance(config, dict):
            return config
        return config.get_dictionary()


class GenotypeConverter(BaseConverter):
    """Converter for the DARTS genotype."""
    
    def __init__(self, name="genotype"):
        super().__init__(name)

    def convert(self, config):
        config_dict = convert_genotype_to_config(config)
        return config_dict


class BANANASConverter(BaseConverter):
    """BANANAS representation is like the DARTS genotype, but the tuples are inverted (first node, then operation)."""

    def __init__(self, name="BANANAS"):
        super().__init__(name)

    def convert(self, config):
        # Convert to genotype
        normal = [(OPS[op], node) for node, op in bananas_arch[0]]
        reduction = [(OPS[op], node) for node, op in bananas_arch[1]]
        concat = list(range(2, 6))
        genotype = Genotype(normal=normal, reduce=reduction, normal_concat=concat, reduce_concat=concat)
        
        # Convert genotype to configspace dictionary
        config_dict = convert_genotype_to_config(genotype)
        return config_dict


CONVERTER_DICT = {
        "configspace" : ConfigspaceInstanceConverter,
        "genotype" : GenotypeConverter,
        "BANANAS" : BANANASConverter
        }



from collections import namedtuple

from ConfigSpace.read_and_write import json as cs_json

import nasbench301 as nb

# Load the performance surrogate model
#NOTE: Loading the ensemble will set the seed to the same as used during training (logged in the model_configs.json)
print("==> Loading performance surrogate model...")
ensemble_dir_performance = "/path/to/nasbench301_models_v0.9/xgb_v0.9"
performance_model = nb.load_ensemble(ensemble_dir_performance)

# Load the runtime surrogate model
print("==> Loading runtime surrogate model...")
ensemble_dir_runtime = "/path/to/nasbench301_models_v0.9/lgb_runtime_v0.9"
runtime_model = nb.load_ensemble(ensemble_dir_runtime)

# Option 1: Create a DARTS genotype
print("==> Creating test configs...")
Genotype = namedtuple('Genotype', 'normal normal_concat reduce reduce_concat')
genotype_config = Genotype(
        normal=[('sep_conv_3x3', 0), ('sep_conv_3x3', 1), ('sep_conv_3x3', 0), ('sep_conv_3x3', 1), ('sep_conv_3x3', 1), ('skip_connect', 0), ('skip_connect', 0), ('dil_conv_3x3', 2)],
        normal_concat=[2, 3, 4, 5],
        reduce=[('max_pool_3x3', 0), ('max_pool_3x3', 1), ('skip_connect', 2), ('max_pool_3x3', 1), ('max_pool_3x3', 0), ('skip_connect', 2), ('skip_connect', 2), ('max_pool_3x3', 1)],
        reduce_concat=[2, 3, 4, 5]
        )

# Option 2: Sample from a ConfigSpace
with open("configspace.json", "r") as f:
    json_string = f.read()
    configspace = cs_json.read(json_string)
configspace_config = configspace.sample_configuration()

# Predict
print("==> Predict runtime and performance...")
prediction_genotype = performance_model.predict(config=genotype_config, representation="genotype", with_noise=True)
prediction_configspace = performance_model.predict(config=configspace_config, representation="configspace", with_noise=True)

runtime_genotype = runtime_model.predict(config=genotype_config, representation="genotype")
runtime_configspace = runtime_model.predict(config=configspace_config, representation="configspace")

print("Genotype architecture performance: %f, runtime %f" %(prediction_genotype, runtime_genotype))
print("Configspace architecture performance: %f, runtime %f" %(prediction_configspace, runtime_configspace))



import os
import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

requirements = []
with open('requirements.txt', 'r') as f:
    for line in f:
        requirements.append(line.strip())

#optional_requirements = []
#with open('optional-requirements.txt', 'r') as f:
#    for line in f:
#        optional_requirements.append(line.strip())

setuptools.setup(
    name="nasbench301",
    version="0.2",
    author="AutoML Freiburg",
    author_email="zimmerl@informatik.uni-freiburg.de",
    description=("A surrogate benchmark for neural architecture search"),
    long_description=long_description,
    url="https://github.com/automl/nasbench301",
    long_description_content_type="text/markdown",
    license="3-clause BSD",
    keywords="machine learning"
             "optimization tuning neural architecture deep learning",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
    ],
	python_requires='>=3',
    platforms=['Linux'],
    #install_requires=requirements,
    include_package_data=True
#    extras_require=optional_requirements
)
